@@ -35,10 +35,10 @@
 #include "vscan.h"
 #include "cache.h"
 #include "os.h"
+#include "ssl.h"
 
 int extract_clientcmd(sstr * buf, sstr ** cmd, sstr ** arg);
 int extract_servercmd(sstr * buf, int *code, sstr ** msg);
-void do_linecontrol(sstr * buf, int reply);
 static session_info *init_info(int fd, struct sockaddr_in source);
 static void connect_to_server(void);
 
@@ -53,14 +53,11 @@ void init_session(int fd, struct sockaddr_in source)
 	write_log(INFO, "Connect from %s",
 		  sstr_buf(addr2name(info->client_control.address.sin_addr)));
 
-	/*FIXME have a login function which deals with ntp and cache*/
+	/*FIXME have a login function which deals with ntp and cache */
 	ccp_changedest();
-	ntp_changedest(); 
+	ntp_changedest();
 	info->final_server_address = info->server_control.address;
 
-	if(config.ftpproxy.sin_addr.s_addr)
-		info->server_control.address = config.ftpproxy;
-
 	write_log(INFO, "... to %s(%s)",
 		  inet_ntoa(info->final_server_address.sin_addr),
 		  sstr_buf(info->server_name));
@@ -74,11 +71,11 @@ void init_session(int fd, struct sockaddr_in source)
 
 static void connect_to_server(void)
 {
-	/*Check for loops. Won't work if Listen undefined*/
+	/*Check for loops. Won't work if Listen undefined */
 	if(info->server_control.address.sin_addr.s_addr
 	   == config.listen_address.sin_addr.s_addr &&
 	   info->server_control.address.sin_port
-	   == config.listen_address.sin_port) 
+	   == config.listen_address.sin_port)
 		die(ERROR, "Attempt to connect to self. "
 		    "Do you need to set DoNTP to yes?",
 		    421, "Proxy tried to loop. Closing connection", 0);
@@ -93,40 +90,49 @@ static void connect_to_server(void)
 			    "Frox unable to determine destination address. "
 			    "Try commenting out NTPAddress",
 			    501, "Unable to contact server", 0);
-		else die(ERROR, "Frox unable to determine detination address. "
-			    "See FAQ for troubleshooting.",
-			    501, "Unable to contact server", 0);
+		else
+			die(ERROR,
+			    "Frox unable to determine detination address. "
+			    "See FAQ for troubleshooting.", 501,
+			    "Unable to contact server", 0);
 	}
 
-	resolve_addr(&info->final_server_address.sin_addr,
-		     info->server_name);
+	resolve_addr(&info->final_server_address.sin_addr, info->server_name);
 
-	if (!config_connectionok(&info->client_control.address,
-				 &info->final_server_address))
+	if(!config_connectionok(&info->client_control.address,
+				&info->final_server_address,
+				info->username ? sstr_buf(info->
+							  username) : 0))
 		die(ATTACK, "Denied by ACLs.", 501, "Connection denied. Bye",
 		    0);
 
+	if(config.ftpproxy.sin_addr.s_addr)
+		info->server_control.address = config.ftpproxy;
+
 	write_log(VERBOSE, "Connecting to server...");
 
 	info->server_control.fd =
 		connect_to_socket(&info->server_control.address,
-				&config.tcpoutaddr,
-				config.contports);
+				  &config.tcpoutaddr, config.contports);
 
-	if (info->server_control.fd == -1)
+	if(info->server_control.fd == -1)
 		die(ERROR, "Connection closed -- unable to contact server",
 		    501, "Proxy unable to contact ftp server", 0);
 
 	write_log(VERBOSE, "     OK");
 
-	if(config.loglevel>=VERBOSE){ /*Save the overhead of DNS lookups*/
-	write_log(VERBOSE, "Apparent address = %s",
-	       sstr_buf(addr2name(info->apparent_server_address.sin_addr)));
-	write_log(VERBOSE, "Real address = %s",
-	       sstr_buf(addr2name(info->final_server_address.sin_addr)));
-	write_log(VERBOSE, "Proxy address = %s",
-	       sstr_buf(addr2name(info->server_control.address.sin_addr)));
+	if(config.loglevel >= VERBOSE) {	/*Save the overhead of DNS lookups */
+		write_log(VERBOSE, "Apparent address = %s",
+			  sstr_buf(addr2name
+				   (info->apparent_server_address.sin_addr)));
+		write_log(VERBOSE, "Real address = %s",
+			  sstr_buf(addr2name
+				   (info->final_server_address.sin_addr)));
+		write_log(VERBOSE, "Proxy address = %s",
+			  sstr_buf(addr2name
+				   (info->server_control.address.sin_addr)));
 	}
+	ssl_init();
 }
 
 /* ------------------------------------------------------------- **
@@ -137,7 +143,8 @@ static session_info *init_info(int fd, struct sockaddr_in source)
 {
 	session_info *i;
 	i = (session_info *) malloc(sizeof(session_info));
-	if (i == NULL) die(ERROR, "Malloc failed", 0, NULL, -1);
+	if(i == NULL)
+		die(ERROR, "Malloc failed", 0, NULL, -1);
 	memset(i, 0, sizeof(session_info));
 
 	i->server_listen = i->client_listen = -1;
@@ -158,9 +165,12 @@ static session_info *init_info(int fd, struct sockaddr_in source)
 
 	i->filename = sstr_init(0);
 
-	if (config.apconv) i->mode = APCONV;
-	else i->mode = ACTIVE;
+	if(config.apconv)
+		i->mode = APCONV;
+	else
+		i->mode = ACTIVE;
 	i->state = NEITHER;
+	i->greeting = AWAITED;
 
 	i->client_control.address = source;
 	i->client_control.fd = fd;
@@ -180,9 +190,11 @@ void run_proxy()
 
 	do {
 		i = get_control_line(GET_BOTH);
-		if (i & GET_CLNT) client_control_forward();
-		if (i & GET_SRVR) server_control_forward();
-	} while (TRUE);
+		if(i & GET_CLNT)
+			client_control_forward();
+		if(i & GET_SRVR)
+			server_control_forward();
+	} while(TRUE);
 }
 
 /* ------------------------------------------------------------- **
@@ -193,11 +205,11 @@ void client_control_forward()
 	int i;
 	sstr *cmd, *arg;
 
-	while ((i =
-		extract_clientcmd(info->client_control.buf, &cmd, &arg)) == 0)
+	while((i =
+	       extract_clientcmd(info->client_control.buf, &cmd, &arg)) == 0)
 		parse_client_cmd(cmd, arg);
 
-	if (i < 0)
+	if(i < 0)
 		die(ATTACK,
 		    "Client is sending us a badly formed control stream.",
 		    421, "You are sending me rubbish. Goodbye.", -1);
@@ -211,15 +223,17 @@ void server_control_forward()
 	int i, code;
 	sstr *arg;
 
-	while ((i =
-		extract_servercmd(info->server_control.buf, &code, &arg)) == 0)
-		if (!cache_parsed_reply(code, arg) && 
-		    !vscan_parsed_reply(code, arg))
+	while((i =
+	       extract_servercmd(info->server_control.buf, &code, &arg)) == 0)
+		if(!ssl_parsed_reply(code, arg) &&	/*Always returns 0 */
+		   !cache_parsed_reply(code, arg) &&
+		   !vscan_parsed_reply(code, arg))
 			send_message(code, arg);
 
-	if (i < 0)
+	if(i < 0)
 		die(ATTACK,
-		    "Server is sending us a badly formed control stream.", 421,
+		    "Server is sending us a badly formed control stream.",
+		    421,
 		    "FTP server is sending you rubbish! Closing connection.",
 		    -1);
 }
@@ -235,13 +249,15 @@ void parse_client_cmd(sstr * cmd, sstr * arg)
 	struct cmd_struct *p;
 	int i;
 
-	if(!ccp_allowcmd(cmd, arg)) return;
-	sstr_cpy(info->last_command,cmd);
+	if(!ccp_allowcmd(cmd, arg))
+		return;
+	sstr_cpy(info->last_command, cmd);
 
-	for(i=0;i<NUM_CMD_ARRAYS;i++) {
-		if(!info->cmd_arrays[i]) continue;
-		for (p = info->cmd_arrays[i]; p && p->cmd; p++) {
-			if (!*p->name || !sstr_casecmp2(cmd, p->name)) {
+	for(i = 0; i < NUM_CMD_ARRAYS; i++) {
+		if(!info->cmd_arrays[i])
+			continue;
+		for(p = info->cmd_arrays[i]; p && p->cmd; p++) {
+			if(!*p->name || !sstr_casecmp2(cmd, p->name)) {
 				p->cmd(cmd, arg);
 				return;
 			}
@@ -258,8 +274,9 @@ void get_command(sstr ** cmd, sstr ** arg)
 {
 	int i;
 
-	while ((i = extract_clientcmd(info->client_control.buf, cmd, arg)) == 1)
-		if (get_control_line(GET_CLNT) <= 0)
+	while((i =
+	       extract_clientcmd(info->client_control.buf, cmd, arg)) == 1)
+		if(get_control_line(GET_CLNT) <= 0)
 			die(ERROR, "Arrrghh. Shouldn't be here", 0, NULL, -1);
 }
 
@@ -269,21 +286,27 @@ void get_command(sstr ** cmd, sstr ** arg)
 void get_message(int *code, sstr ** msg)
 {
 	do {
-		switch(extract_servercmd(info->server_control.buf, code, msg)){
+		switch (extract_servercmd
+			(info->server_control.buf, code, msg)) {
 		case -1:
 			die(ATTACK,
-		         "Server is sending us a badly formed control stream.",
-		          421, "FTP server is sending you rubbish!"
-			  "Closing connection.", -1);
+			    "Server is sending us a badly formed control stream.",
+			    421, "FTP server is sending you rubbish!"
+			    "Closing connection.", -1);
 		case 0:
-			if(*code > 0) return;
+#ifdef DEBUG
+			fprintf(stderr, "  s: \033[32m%d %s\033[37m\n",
+				*code, msg ? sstr_buf(*msg) : "");
+#endif
+			if(*code > 0)
+				return;
 			continue;
 		case 1:
-			if (get_control_line(GET_SRVR) <= 0)
+			if(get_control_line(GET_SRVR) <= 0)
 				die(ERROR, "Arrghh - shoudln't be here",
 				    0, NULL, -1);
 		}
-	}while(TRUE);
+	} while(TRUE);
 }
 
 
@@ -318,7 +341,7 @@ void send_command(sstr * cmd, sstr * arg)
 	buf = sstr_init(MAX_LINE_LEN + 10);
 
 	sstr_cat(buf, cmd);
-	if (sstr_len(arg) != 0) {
+	if(sstr_len(arg) != 0) {
 		sstr_ncat2(buf, " ", 1);
 		sstr_cat(buf, arg);
 	}
@@ -329,7 +352,10 @@ void send_command(sstr * cmd, sstr * arg)
 #else
 	write_log(VERBOSE, "  C: %s", sstr_buf(buf));
 #endif
-	sstr_write(info->server_control.fd, buf, 0);
+	if(info->ssl_sc)
+		ssl_write(info->ssl_sc, buf);
+	else
+		sstr_write(info->server_control.fd, buf, 0);
 
 	sstr_free(buf);
 }
@@ -340,11 +366,12 @@ void send_command(sstr * cmd, sstr * arg)
 void send_message(int code, sstr * msg)
 {
 	sstr *buf;
-	if(!ccp_allowmsg(&code, msg)) return;
+	if(!ccp_allowmsg(&code, msg))
+		return;
 
 	buf = sstr_init(MAX_LINE_LEN + 10);
 
-	if (code != 0)
+	if(code != 0)
 		sstr_apprintf(buf, "%d%c", abs(code), code > 0 ? ' ' : '-');
 	sstr_cat(buf, msg);
 	sstr_ncat2(buf, "\r\n", 2);
@@ -358,6 +385,16 @@ void send_message(int code, sstr * msg)
 	sstr_free(buf);
 }
 
+int read_srvrctrl_data(void)
+{
+	if(info->ssl_sc)
+		return ssl_append_read(info->ssl_sc,
+				       info->server_control.buf, 0);
+	else
+		return sstr_append_read(info->server_control.fd,
+					info->server_control.buf, 0);
+}
+
 /* ------------------------------------------------------------- **
 ** Central select bit. Deals with data connection
 ** forwarding/listening, and quits on ctrl connection close. Once
@@ -377,10 +414,10 @@ int get_control_line(int which)
 	fd_set reads, writes;
 
 	do {
-		i=setup_fds(&reads, &writes);
+		i = setup_fds(&reads, &writes);
 		alarm(config.timeout);
-		if (select(i+1, &reads, &writes, NULL, NULL) == -1) {
-			if (errno == EINTR)
+		if(select(i + 1, &reads, &writes, NULL, NULL) == -1) {
+			if(errno == EINTR)
 				continue;
 			debug_perr("select");
 			die(0, NULL, 0, NULL, -1);
@@ -388,101 +425,47 @@ int get_control_line(int which)
 
 		do_dataforward(&reads, &writes);
 
-		if (FD_ISSET(info->client_control.fd, &reads)) {
+		if(FD_ISSET(info->client_control.fd, &reads)) {
 			i = sstr_append_read(info->client_control.fd,
 					     info->client_control.buf, 0);
-			if (i == 0)
+			if(i == 0)
 				die(INFO, "Client closed connection",
 				    0, NULL, 0);
-			if (i < 0)
+			if(i < 0)
 				die(ATTACK,
-				     "Client flooding control connection", 421,
-				     "You're sending rubbish. Goodbye", -1);
-			if (sstr_hasline(info->client_control.buf))
+				    "Client flooding control connection", 421,
+				    "You're sending rubbish. Goodbye", -1);
+			if(sstr_hasline(info->client_control.buf))
 				ret |= GET_CLNT;
 		}
 
-		if (info->server_control.fd != -1 &&
-		    FD_ISSET(info->server_control.fd, &reads)) {
-			i = sstr_append_read(info->server_control.fd,
-					     info->server_control.buf, 0);
-			if (i == 0) {
+		if(info->server_control.fd != -1 &&
+		   FD_ISSET(info->server_control.fd, &reads)) {
+			i = read_srvrctrl_data();
+			if(i == 0) {
 				if(!cache_transferring())
 					die(ERROR,
-					"Server closed the control connection",
-					0, NULL, 0);
+					    "Server closed the control connection",
+					    0, NULL, 0);
 				else {
 					rclose(&info->server_control.fd);
-					write_log(INFO, "Server closed connection. Keeping going until cache done");
+					write_log(INFO,
+						  "Server closed connection. Keeping going until cache done");
 				}
 			}
-			if (i < 0)
+			if(i < 0)
 				die(ATTACK,
 				    "Server flooding the control connection",
 				    421, "Server is sending rubbish."
 				    "Closing connection", -1);
-			if (sstr_hasline(info->server_control.buf))
+			if(sstr_hasline(info->server_control.buf))
 				ret |= GET_SRVR;
 		}
 
-	} while (!(ret & which));
+	} while(!(ret & which));
 	return (ret);
 }
 
-/* ------------------------------------------------------------- **
-** Process buf to remove telnet line control stuff, and reply on fd
-** reply. FIXME this is quite ugly.
-** ------------------------------------------------------------- */
-void do_linecontrol(sstr * buf, int reply)
-{
-	int i;
-	sstr *tmp;
-
-	tmp = sstr_init(3);
-
-	while ((i = sstr_chr(buf, IAC)) != -1) {
-		switch (sstr_getchar(buf, i + 1) & 255) {
-		case WILL:
-			if (sstr_split(buf, tmp, i, 3) == -1)
-				goto rubbish;
-			sstr_setchar(tmp, 1, DONT);
-			send(reply, sstr_buf(tmp), 3, MSG_OOB);
-			break;
-		case DO:
-			if (sstr_split(buf, tmp, i, 3) == -1)
-				goto rubbish;
-			sstr_setchar(tmp, 1, WONT);
-			send(reply, sstr_buf(tmp), 3, MSG_OOB);
-			break;
-		case WONT:
-		case DONT:
-			if (sstr_split(buf, NULL, i, 3) == -1)
-				goto rubbish;
-			break;
-		case IAC:
-			if (sstr_split(buf, NULL, i, 1) == -1)
-				goto rubbish;
-			break;
-		case IP:
-		case DM:
-			if (sstr_split(buf, NULL, i, 2) == -1)
-				goto rubbish;
-			break;
-		default:
-			if (sstr_split(buf, NULL, i, 2) == -1)
-				goto rubbish;
-			break;
-		}
-	}
-	sstr_free(tmp);
-	return;
-
-rubbish:
-	die(ATTACK,
-	    "Client sending dodgy telnet control characters",
-	    421, "You are sending me rubbish. Closing connection", -1);
-}
-
 /***************************************************************************
  *
  * Functions which read the raw control stream --- be careful
@@ -502,16 +485,16 @@ int extract_clientcmd(sstr * buf, sstr ** pcmd, sstr ** parg)
 {
 	static sstr *cmd = NULL, *arg = NULL;
 
-	if (!cmd)
+	if(!cmd)
 		cmd = sstr_init(4);
-	if (!arg)
+	if(!arg)
 		arg = sstr_init(MAX_LINE_LEN);
-	if(pcmd) *pcmd = cmd;
-	if(parg) *parg = arg;
+	if(pcmd)
+		*pcmd = cmd;
+	if(parg)
+		*parg = arg;
 
-	do_linecontrol(buf, info->client_control.fd);
-
-	if (!sstr_hasline(buf))
+	if(!sstr_hasline(buf))
 		return (1);
 	switch (sstr_token(buf, cmd, " \t\n\r", 0)) {
 	case -1:		/*Token doesn't fit in cmd */
@@ -524,7 +507,8 @@ int extract_clientcmd(sstr * buf, sstr ** pcmd, sstr ** parg)
 		sstr_empty(arg);
 	}
 
-	if(!config.nonasciiok) sstr_makeprintable(arg, '?');
+	if(!config.nonasciiok)
+		sstr_makeprintable(arg, '?');
 
 	return (0);
 }
@@ -539,25 +523,24 @@ int extract_servercmd(sstr * buf, int *code, sstr ** pmsg)
 	static int multiline = 0;
 	static sstr *scode = NULL, *msg = NULL;
 
-	if (!scode)
+	if(!scode)
 		scode = sstr_init(4);
-	if (!msg)
+	if(!msg)
 		msg = sstr_init(MAX_LINE_LEN);
-	if(pmsg) *pmsg = msg;
+	if(pmsg)
+		*pmsg = msg;
 
-	do_linecontrol(buf, info->client_control.fd);
-
-	if (!sstr_hasline(buf))
+	if(!sstr_hasline(buf))
 		return (1);
 
-	if (sstr_token(buf, msg, "\r\n", 0) == -1)
+	if(sstr_token(buf, msg, "\r\n", 0) == -1)
 		return (-1);
 
-	if (!multiline ||
-	    (sstr_atoi(msg) == multiline && sstr_getchar(msg, 3) == ' ')) {
+	if(!multiline ||
+	   (sstr_atoi(msg) == multiline && sstr_getchar(msg, 3) == ' ')) {
 		multiline = 0;
 
-		if (sstr_split(msg, scode, 0, 4) == -1)
+		if(sstr_split(msg, scode, 0, 4) == -1)
 			return (-1);
 		*code = sstr_atoi(scode);
 
@@ -574,12 +557,8 @@ int extract_servercmd(sstr * buf, int *code, sstr ** pmsg)
 		*code = 0;
 	}
 
-	if(!config.nonasciiok) sstr_makeprintable(msg, '?');
+	if(!config.nonasciiok)
+		sstr_makeprintable(msg, '?');
 
 	return (0);
 }
-
-
-
-
-
